import cors from '@fastify/cors';
import { Type, TypeBoxTypeProvider } from '@fastify/type-provider-typebox';
import Fastify, { FastifyReply } from 'fastify';
import { nanoid } from 'nanoid';
import { allocateBuffer, IMAGES } from 'runner';
import { Pull, Push } from 'zeromq';

const sender = new Push();
await sender.bind(`tcp://127.0.0.1:5557`);
const receiver = new Pull();
await receiver.bind(`tcp://127.0.0.1:5558`);

const clients: Record<string, FastifyReply> = {};
const generateId = () => nanoid(32);

const fastify = Fastify({
  logger: true,
}).withTypeProvider<TypeBoxTypeProvider>();
await fastify.register(cors, {
  origin: process.env.ALLOW_ORIGIN || '*',
});

fastify.post('/run', {
  schema: {
    body: Type.Object({
      code: Type.String(),
      language: Type.String(),
    }),
  },
}, (req, reply) => {
  const { code, language } = req.body;
  const jobId = generateId();
  const buffer = allocateBuffer(jobId, code, IMAGES.moshell);
  reply.raw.writeHead(200, {
    'Content-Type': 'text/event-stream',
    Connection: 'keep-alive',
    'Cache-Control': 'no-cache',
    'Access-Control-Allow-Origin': process.env.ALLOW_ORIGIN || '*',
  });
  reply.raw.on('close', () => {
    delete clients[jobId];
  });
  sender.send(buffer).then(() => {
    reply.raw.write('event: connected\n');
    reply.raw.write(`data: ${jobId}\n`);
    reply.raw.write('id: 0\n\n');
  });
  clients[jobId] = reply;
});

async function forwardOutput() {
  for await (const [buff] of receiver) {
    const messageType = buff.readInt8();
    const jobId = buff.subarray(1, 33).toString();
    const reply = clients[jobId];
    if (!reply) {
      continue;
    }
    console.debug(`Forwarding message type ${messageType} for job ${jobId}`);
    const raw = reply.raw;
    switch (messageType) {
      case 1:
      case 2:
        const text = encodeURIComponent(buff.subarray(37).toString());
        raw.write('event: message\n');
        if (messageType === 1) {
          raw.write('id: stdout\n');
        } else {
          raw.write('id: stderr\n');
        }
        raw.write(`data: ${text}\n\n`);
        break;
      case 3:
        const exitCode = buff.readUint32BE(33);
        raw.write('event: message\n');
        raw.write('id: exit\n');
        raw.write(`data: ${exitCode}\n\n`);
        raw.end();
        break;
      default:
        console.error('Unknown message type', messageType);
    }
  }
}

await Promise.all([fastify.listen({ port: 3000 }), forwardOutput()]);
